/*******************************************************************************
 * File Name	: kb.c
 * Project	: Keycard
 * Date		: 2014/01/09
 * Version	: 1.0
 * Target MCU	: ATMEGA16U2
 * Tool Chain	: Atmel AVR Studio 4.19 730, avr-gcc, avr-libc
 * Author	: "Detlef Mueller" <detlef@gmail.com>
 * References	:
 * Device	: TechKeys Business Card
 *
 * Release Notes:
 *
 * Pinout ATmega16U2
 * =================
 *
 * PB0 --               PC0 --           PD0 Conn 3
 * PB1 System LED       PC1 --           PD1 --
 * PB2 Right LED        PC2 Conn 2       PD2 --
 * PB3 Middle LED                        PD3 Conn 5
 * PB4 Left LED         PC4 --           PD4 Conn 6
 * PB5 Right Key        PC5 --           PD5 Conn 7
 * PB6 Middle Key       PC6 --           PD6 Conn 4
 * PB7 Left Key         PC7 --           PD7 --      
 *
 * $Id$
 ******************************************************************************/

#define	__FIND_BL_ADDR

#include <includes.h>
#include "usb_cfg.h"
#include <u_des_kb.h>
#include <u_hid_kb.h>
#include <kbusages.h>
#include <kbmacro.h>

#include "kb.h"

//------------------------------------------------------------------------------
//******************************************************************************
//------------------------------------------------------------------------------

#define	DDO( b )	_B1(b)		/* data direction: out */
#define	DDI( b )	_B0(b)		/* data direction: in  */
#define	PU1( b )	_B1(b)		/* pull up on (if DDI) */
#define	PU0( b )	_B0(b)		/* pull up off */

// Port B

#define PBPU		(PU1( PB7) | PU1( PB6) | PU1( PB5) | _B1( PB4) | \
			 _B1( PB3) | _B1( PB2) | _B1( PB1) | PU1( PB0))
#define DDB		(DDI(DDB7) | DDI(DDB6) | DDI(DDB5) | DDO(DDB4) | \
			 DDO(DDB3) | DDO(DDB2) | DDO(DDB1) | DDI(DDB0))
// Port C

#define PCPU		(PU1( PC7) | PU1( PC6) | PU1( PC5) | PU1( PC4) | \
			             PU1( PC2) | PU1( PC1) | PU1( PC0))
#define DDC		(DDI(DDC7) | DDI(DDC6) | DDI(DDC5) | DDI(DDC4) | \
			             DDI(DDC2) | DDI(DDC1) | DDI(DDC0))
// Port D

#define PDPU		(PU1( PD7) | PU1( PD6) | PU1( PD5) | PU1( PD4) | \
			 PU1( PD3) | PU1( PD2) | PU1( PD1) | PU1( PD0))
#define DDD		(DDI(DDD7) | DDI(DDD6) | DDI(DDD5) | DDI(DDD4) | \
			 DDI(DDD3) | DDI(DDD2) | DDI(DDD1) | DDI(DDD0))

//------------------------------------------------------------------------------
// 3% Key & LED definitions

#define	KEY_IN_P		PINB

#define	DEBOUNCE_MS		4	/* Cherry says 5ms debounce time, this is 2ms */

// Number of keys on the board, 3 on the TechKeys Business Card, each w/ a LED

#define	KEY_COUNT		3
#define	LED_COUNT		3

// Key bits and masks

#define bKEY_RGT		0			/* PB5 */
#define bKEY_MID		1			/* PB6 */
#define bKEY_LFT		2			/* PB7 */

#define	mKEYS			0b00000111
#define	mKEYS_P			0b11100000

// LED bits and masks

#define	bLED_RGT		PB2
#define	bLED_MID		PB3
#define	bLED_LFT		PB4

#define	mLED_RGT		_BV(bLED_RGT)
#define	mLED_MID		_BV(bLED_MID)
#define	mLED_LFT		_BV(bLED_LFT)

#define	mLEDS			(mLED_RGT | mLED_MID | mLED_LFT)

//------------------------------------------------------------------------------

// Bit definitions for the keyboard LED OUT report

#define	bLED_NUM		0
#define	bLED_CAPS		1
#define	bLED_SCR		2
#define	bLED_COMPOSE		3
#define	bLED_KANA		4

// Bit definitions for the keyboard modifiers in the IN report

#define	bLFT_CTRL		0
#define	bLFT_SHFT		1
#define	bLFT_ALT		2
#define	bLFT_GUI		3
#define	bRGT_CTRL		4
#define	bRGT_SHFT		5
#define	bRGT_ALT		6
#define	bRGT_GUI		7

//------------------------------------------------------------------------------
//******************************************************************************
//------------------------------------------------------------------------------

volatile uint8_t
    alarm ;

//------------------------------------------------------------------------------

static uint8_t
    demo_off,				// No demo mode flag (can be irritating..)
    led_r_pat,				// Right LED PWM pattern
    led_m_pat,				// Middle LED PWM pattern
    led_l_pat ;				// Left LED PWM pattern

//------------------------------------------------------------------------------
//******************************************************************************
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Misc. helper functions
//------------------------------------------------------------------------------

// Reboot into bootloader.

static void FA_NORETURN( bootloader ) ( void )
{
    TIMSK0 = 0 ;			// Kill timer INT
    UDCON  = _B1(DETACH) ;		// Detach from USB

    cli() ;				// Disable interrupts

    USBCON = _B0(USBE) | _B1(FRZCLK) ;	// Stop USB module
    clr_bit( PLLCSR, PLLE ) ;		// Stop PLL

    wdt_disable() ;			// Leash watchdog

    // Disable peripherals

    DDRB   = 0 ; PORTB  = 0 ;
    DDRC   = 0 ; PORTC  = 0 ;
    DDRD   = 0 ; PORTD  = 0 ;

    TCCR0A = 0 ; TCCR0B = 0 ;
    OCR0A  = 0 ;

    PRR0   = 0 ;			// Flip needs timer 1..

    // 15ms to make sure detach is recognized

    _delay_loop_2( (uint16_t)MS2TM( 15, 4 ) ) ;

    jmp_bootloader() ;			// Off we go
}

//------------------------------------------------------------------------------
// LED related functions
//------------------------------------------------------------------------------

// Maintain Sleep LED. Note: called every 1ms (system clock throttled)

#define	SLEEP_OFF_DELAY		1				/* 1ms on */
#define	SLEEP_ON_DELAY		(3000 - SLEEP_OFF_DELAY)	/* once w/in 3s */

void sleep_led ( uint8_t reset )
{
    static uint16_t
	timer ;

    if ( reset )
    {
	LED_off( bLED_SYS ) ;
	timer = 0 ;

	return ;
    }

    if ( ! timer-- )			// Timer expired
    {
	if ( LED_sts( bLED_SYS ) )	// LED on ?
	{
	    LED_off( bLED_SYS ) ;
	    timer = SLEEP_ON_DELAY ;
	}
	else
	{
	    LED_on( bLED_SYS ) ;
	    timer = SLEEP_OFF_DELAY ;
	}
    }
}

//------------------------------------------------------------------------------

// Set the PWM pattern (ie. dim level) for the specified LED's

static void set_leds ( uint8_t leds, uint8_t lvl )
{
    static const uint8_t
	pattern[] PROGMEM =
	{
	    0b00000000,	//   0%
	    0b00000001,	//  12.5%
	    0b10001000,	//  25%
	    0b01001010,	//  37.5%
	    0b01010101,	//  50%
	    0b11011010,	//  62.5%
	    0b01110111,	//  75
	    0b11111110,	//  82.5%
	    0b11111111	// 100%
	} ;

    static uint8_t
	r_pat, m_pat, l_pat ;

//    if ( lvl >= ARRSZ( pattern ) )		// Should never happen
//	return ;

    lvl = pgm_read_byte( pattern + lvl ) ;

    if ( (leds & mLED_RGT) && r_pat != lvl )
    {
	r_pat     = lvl ;
	led_r_pat = lvl ;
    }

    if ( (leds & mLED_MID) && m_pat != lvl )
    {
	m_pat     = lvl ;
	led_m_pat = lvl ;
    }

    if ( (leds & mLED_LFT) && l_pat != lvl )
    {
	l_pat     = lvl ;
	led_l_pat = lvl ;
    }
}
//------------------------------------------------------------------------------

// Maintain LED PWM

void led_pwm ( void )
{
    uint8_t
	pat ;

    if ( (pat = led_r_pat) & 1 )
	LED_on( bLED_RGT ) ;
    else
	LED_off( bLED_RGT ) ;

    led_r_pat = ror8( pat ) ;

    if ( (pat = led_m_pat) & 1 )
	LED_on( bLED_MID ) ;
    else
	LED_off( bLED_MID ) ;

    led_m_pat = ror8( pat ) ;

    if ( (pat = led_l_pat) & 1 )
	LED_on( bLED_LFT ) ;
    else
	LED_off( bLED_LFT ) ;

    led_l_pat = ror8( pat ) ;
}

//------------------------------------------------------------------------------

#define	DEMO_TIMER		45		/* Change LED's every 45ms */

static void led_demo ( uint8_t reset )
{
    static const uint8_t
	breathe[] PROGMEM =
	{
	   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	   1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 4, 4, 5, 6, 7,
	   8, 7, 6, 5, 4, 4, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1,
	   1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	} ;

    static uint8_t
	timer,
	idx ;

    if ( reset )
    {
	timer = 0 ;
	idx   = 0 ;

	set_leds( mLEDS, 0 ) ;

	return ;
    }

    if ( ! timer-- )
    {
	timer = DEMO_TIMER ;

	set_leds( mLED_LFT, pgm_read_byte( breathe +   idx  ) ) ;
	set_leds( mLED_MID, pgm_read_byte( breathe + ((idx +  7) % ARRSZ( breathe )) ) ) ;
	set_leds( mLED_RGT, pgm_read_byte( breathe + ((idx + 14) % ARRSZ( breathe )) ) ) ;

	if ( ++idx >= ARRSZ( breathe ) )
	    idx = 0 ;
    }
}

//------------------------------------------------------------------------------

// Maintain LED's

#define	DEMO_DELAY		10000		/* Start demo after 10s w/o LEDs */
#define	BOOT_DELAY		15000		/* Start boot-loader delay, 15s  */
#define	LED_TIMER		1		/* 500us */

void maint_leds ( uint8_t reset )
{
    static uint16_t
	bload,				// Bootloader call timer
	demo ;				// Demo start timer

    static uint8_t
	timer ;

    uint8_t
	mLeds ;

    if ( reset )
    {
	bload = BOOT_DELAY ;
	demo  = DEMO_DELAY ;
	timer = 0 ;

	set_leds( mLEDS, 0 ) ;
	led_pwm() ;			// Make sure LED's are off

	return ;
    }

    if ( ! timer-- )			// Maintain LED's every .5ms * timer
    {
	timer = LED_TIMER ;

	if ( (kbd_leds & (_BV(bLED_NUM) | _BV(bLED_CAPS) | _BV(bLED_SCR))) )
	{					// LOCK LED(s) active
	    if ( demo != DEMO_DELAY )		// Demo mode was possibly on
	    {
		bload = BOOT_DELAY ;
		demo  = DEMO_DELAY ;		// Reset demo & bload timers
	    }
						// Scroll- + Num-Lock on
	    if ( kbd_leds == (_BV(bLED_NUM) | _BV(bLED_SCR)) )
		if ( ! --bload )
		    bootloader() ;

	    mLeds = 0 ;

	    if ( kbd_leds & _BV(bLED_NUM) )
		mLeds |= mLED_LFT ;

	    if ( kbd_leds & _BV(bLED_CAPS) )
		mLeds |= mLED_MID ;

	    if ( kbd_leds & _BV(bLED_SCR) )
		mLeds |= mLED_RGT ;

	    set_leds( mLeds, 3 ) ;		// 37.5%
	    set_leds( mLeds ^ mLEDS, 0 ) ;
	}
	else				// No LED active
	{
	    if ( demo )			// Waiting for demo display
	    {
		if ( demo_off )
		{
		    LED_on( bLED_SYS ) ;

		    if ( demo == DEMO_DELAY )	// First run after LED's off
		    {
			set_leds( mLEDS, 0 ) ;	// Set all LED's off
			--demo ;
		    }

		    LED_off( bLED_SYS ) ;
		}
		else
		{
		    if ( demo == DEMO_DELAY )	// First run after LED's off
			set_leds( mLEDS, 0 ) ;	// Set all LED's off

		    if ( ! --demo )		// Demo enable on next run
			led_demo( SIG_RESET ) ;	// Reset demo routine
		}
	    }
	    else
		led_demo( SIG_MAINT ) ;
	}
    }
}

//------------------------------------------------------------------------------
// Key related functions
//------------------------------------------------------------------------------

// Check for any key down in sleep mode

uint8_t wakeup ( uint8_t reset )
{
    static uint8_t
	db ;				// Debounce timer

    uint8_t
	key,
	ret = FALSE ;

    if ( reset )
    {
	db = 0 ;			// Not debouncing

	return ( FALSE ) ;
    }

    key = KEY_IN_P & mKEYS_P ;

    if ( ! db )				// Not debouncing
    {
	if ( key != mKEYS_P )		// Any key down
	    db = DEBOUNCE_MS ;		// start debouncing
    }
    else				// Debouncing
    {
	if ( key == mKEYS_P )		// No key down anymore
	    db = 0 ;			// false alarm
	else				// Any key still own
	if ( ! --db )			// Debounce period expired
	    ret = TRUE ;
    }

    return ( ret ) ;
}

//------------------------------------------------------------------------------

// Build the keyboard report

#define REP_KB		0x0000		/* Keyboard */
#define REP_M		0x1000		/* Macro */
#define REP_SC		0x2000		/* System Control */
#define REP_CC		0x4000		/* Consumer Control */
#define	REP_DEL		0x8000		/* Delay (for macros) */

static uint8_t build_kbd_report ( uint8_t keys )
{
  #define U_Macro_1	0x0010
  #define U_Macro_2	0x0020
  #define U_Macro_3	0x0030

  #if __TECHKEYS
    static const uint16_t
	key_usages[KEY_COUNT] PROGMEM =	// Key-to-usage matrix
	{
	    uM(U_Macro_3),		// PB5 Right key
	    uM(U_Macro_2),		// PB6 Middle key
	    uM(U_Macro_1)		// PB7 Left key
	},
	key_usagesS[KEY_COUNT] PROGMEM = // Key-to-usage matrix if SCR lock is set
	{
	    uCC(U_NextTrack),		// PB5 Right key
	    uCC(U_PlayPause),		// PB6 Middle key
	    uCC(U_PrevTrack)		// PB7 Left key
	},
	key_usagesN[KEY_COUNT] PROGMEM = // Key-to-usage matrix if NUM lock is set
	{
	    uCC(U_VolUp),		// PB5 Right key
	    uCC(U_Mute),		// PB6 Middle key
	    uCC(U_VolDwn)		// PB7 Left key
	} ;
  #else
    static const uint16_t
	key_usages[KEY_COUNT] PROGMEM =	// Key-to-usage matrix
	{
	    uKB(U_ScrollLock),		// PB5 Right key
	    uCC(U_AL_Email),		// PB6 Middle key
	    uCC(U_AC_Home)		// PB7 Left key
	},
	key_usagesS[KEY_COUNT] PROGMEM = // Key-to-usage matrix if SCR lock is set
	{
	    uM(U_Macro_3),		// PB5 Right key
	    uM(U_Macro_2),		// PB6 Middle key
	    uM(U_Macro_1)		// PB7 Left key
	},
	key_usagesN[KEY_COUNT] PROGMEM = // Key-to-usage matrix if NUM lock is set
	{
	    uCC(U_NextTrack),		// PB5 Right key
	    uCC(U_PlayPause),		// PB6 Middle key
	    uCC(U_PrevTrack)		// PB7 Left key
	} ;
  #endif

    static kb_report_t
	last_kbd_rep ;

    static uint16_t
	last_ctrl_rep1,
	last_ctrl_rep2 ;

    uint16_t
	usage ;

    uint8_t
	i,
	ki = 0,
	ret = 0 ;

    const uint16_t
	*pu ;

    // Select key-to-usage table

    if ( kbd_leds == _BV(bLED_NUM) )
	pu = key_usagesN ;
    else
    if ( kbd_leds == _BV(bLED_SCR) )
	pu = key_usagesS ;
    else
	pu = key_usages ;

    if ( keys == mKEYS )		// All keys down, toggle demo mode
	demo_off = ! demo_off ;

    ctrl_report1.key = U_None ;		// Clear System Control Report
    ctrl_report2.key = U_None ;		// Clear Consumer Control Report

    kbd_report.mod   = 0 ;		// Clear modifiers

    for ( i = 0 ; i < KEY_COUNT ; ++i, keys >>= 1 )
    {						// Assemble keyboard report
	if ( (keys & 1) )			// Key down
	{
	    usage = pgm_read_word( pu + i ) ;	// get its usage

	    if ( (usage & REP_M) )		// Macro
		ret |= (usage & 0x00F0) ;
	    else
	    if ( (usage & REP_SC) )		// System Control usage
		ctrl_report1.key = (usage & 0x0FFF) ;
	    else
	    if ( (usage & REP_CC) )		// Consumer Control usage
		ctrl_report2.key = (usage & 0x0FFF) ;
	    else				// Keyboard usage
	    {
		if ( (usage & 0xF0) == 0xE0 )	// Modifier
		    kbd_report.mod |= _bv8( usage & 0x07 ) ;	// Magic.
		else
		    kbd_report.keys[ki++] = (usage & 0xFF) ;
	    }
	}
    }

    // Note: assuming NUM_KEYS <= ARRSZ( kbd_report.keys )

    for ( ; ki < ARRSZ( kbd_report.keys ) ; )
	kbd_report.keys[ki++] = U_None ;

    // Check if a report changed and copy reports

    if ( last_kbd_rep.mod != kbd_report.mod )
    {
	last_kbd_rep.mod = kbd_report.mod ;
	ret |= REP0_CHG ;
    }

    for ( ki = ARRSZ( kbd_report.keys ) ; ki-- ; )
	if ( last_kbd_rep.keys[ki] != kbd_report.keys[ki] )
	{
	    last_kbd_rep.keys[ki] = kbd_report.keys[ki] ;
	    ret |= REP0_CHG ;
	}

    if ( last_ctrl_rep1 != ctrl_report1.key )
    {
	last_ctrl_rep1 = ctrl_report1.key ;
	ret |= REP1_CHG ;
    }

    if ( last_ctrl_rep2 != ctrl_report2.key )
    {
	last_ctrl_rep2 = ctrl_report2.key ;
	ret |= REP2_CHG ;
    }

    return ( ret ) ;
}

//------------------------------------------------------------------------------

// Read and maintain key matrix

#define mKDN			0b00001111	/* 2ms stable */
#define mKUP			0b00010000

uint8_t read_matrix ( uint8_t reset )
{
    static uint8_t
	macro,				// Flag to signal macro playback in progress
	keys ;

    static int8_t
	db_c[KEY_COUNT] ;		// Debounce counters

    uint8_t
	i,
	ret = FALSE,
	key,
	key_msk ;

    int8_t
	*db = db_c ;

    if ( reset )
    {
	for ( i = KEY_COUNT ; i-- ; )		// Clear debounce counters
	    *db++ = 0 ;

	keys = 0 ;				// All keys off

	return ( build_kbd_report( keys ) ) ;	// Clear reports
    }

    if ( macro )			// Playing macro
    {
	ret = play_macro( 0 ) ;

	macro = ret & M_ACTIVE ;
	ret  &= 0x0F ;

	return ( ret ) ;
    }

    key = KEY_IN_P ;			// Active inputs: 0b11100000

    asm volatile
    (
	"swap   %0" "\n\t"
	"lsr    %0" "\n\t"
	: "+r" (key) : "0" (key)
    ) ;					// now 0b00000111

    for ( key_msk = 1, i = KEY_COUNT ; i-- ; ++db, key_msk <<= 1 )
    {
	*db = ((*db & mKDN) << 1) | ! (key & key_msk) ;

	if ( *db == mKUP &&   (keys & key_msk) )
	{
	    keys &= ~key_msk ;
	    ret   = TRUE ;
	}
	else
	if ( *db == mKDN && ! (keys & key_msk) )
	{
	    keys |= key_msk ;
	    ret   = TRUE ;
	}
    }

    if ( ret )				// Key(s) changed
	if ( (ret = build_kbd_report( keys )) & 0xF0 )
	{				// Initialize macro playback
	    ret = play_macro( (ret & 0xF0) >> 4 ) ;

	    macro = ret & M_ACTIVE ;
	    ret  &= 0x0F ;
	}

    return ( ret ) ;
}

//------------------------------------------------------------------------------
// Initialization
//------------------------------------------------------------------------------

// Set up alarm timer

void set_alarm ( void )
{
    ResetTM( 0, 0 ) ;			// Reset T0
    ResetOCRFA( 0 ) ;			// Clear T0 Output Compare A match Flag
    set_bit( TIMSK0, OCIE0A ) ;		// Enable T0 COMPA IRQ
}

//------------------------------------------------------------------------------

// Alarm interrupt handler

ISR( TIMER0_COMPA_vect )
{
    alarm = TRUE ;			// Signal that 500us are up
}

//------------------------------------------------------------------------------

// Delay using T0 (8bit) as timing reference

static void Delay_T0 ( uint8_t time )
{
    set_bit( GTCCR, PSRSYNC ) ;		// reset prescaler

    ResetTM( 0, time ) ;		// set timer

    for ( ; ! TMexp( 0 ) ; )		// wait until timer expired
	;
}

//------------------------------------------------------------------------------

// Initialize the hardware

void hw_init ( void )
{
    uint8_t
	i;

    // Power down..

    ACSR  = _BV(ACD) ;				// analog comp
    PRR0  =					// T0/1, SPI
//	    _BV(PRTIM0) |			// We need T0
	    _BV(PRTIM1) |
	    _BV(PRSPI) ;
    PRR1  =					// USART1
	    _BV(PRUSART1) ;

    clock_prescale_set( clock_div_1 ) ;		// Set clock divider to 1, full speed

    // Initialize ports

    DDRB  = DDB ; PORTB = PBPU ;
    DDRC  = DDC ; PORTC = PCPU ;
    DDRD  = DDD ; PORTD = PDPU ;

    SetTMPS( 0, 1024 ) ;			// Set T1 prescaler to / 1024 for ms delay

    for ( i = 13 ; i-- ; )			// Wait ~200ms to settle in
	Delay_T0( (uint8_t)-MS2TM( 16, 1024 ) ) ;

    // T0 is going to be used for a 500us timebase

    SetTMPS( 0, 64 ) ;				// Set prescaler to / 64

    OCR0A = US2TM( 500, 64 ) ;			// Set output compare reg A to 500us

    set_bit( TCCR0A, WGM01 ) ;			// Set CTC mode
}

//------------------------------------------------------------------------------
